/*********************************************************************
 *
 *                  UART buffered
 *                  A50akord
 *
 * v.1.1 - 31.07.2021, L.Pekny - ATmega*
 * v.1.0 - 15.07.2019, UART, INT, FIFO, L.Pekny
 ********************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
//#include <util/delay.h>
#include "MTimer.h"
#include "UartBuf.h"


UARTblk UART[1];	// Ux=0..0


void UARTopen(BYTE Ux)  // Config UART, clear buf, set interrupts
{
	UART0_CTRLA=(1<<U2X0);		// Enable 1/8 speed
	UART0_CTRLB=(1<<RXCIE0)|(1<<RXEN0)|(1<<TXEN0);
	UART0_CTRLC=(3<<UCSZ00);	// 8b

	UARTinit(Ux);
}


void UARTclose(BYTE Ux)  // Disable UART and interrupts
{
	UART0_CTRLB=0;			
}


BYTE UARTread(BYTE Ux)  // Data? -> read from RXbuf
{
	BYTE value;

	UART0_CTRLB&=~(1<<RXCIE0);			// Dis. Interrupt
	if (UART[Ux].RXrd!=UART[Ux].RXwr)	// Is data?
	  { UART[Ux].RXrd++; UART[Ux].RXrd&=UartCntMask; }
	value=UART[Ux].RXbuf[UART[Ux].RXrd];
	UART0_CTRLB|=(1<<RXCIE0);			// En. Interrupt
	return value;
}


void UARTwrite(BYTE Ux, BYTE value)  // Write to TXbuf, interrupt en
{
	UART0_CTRLB&=~(1<<UDRIE0);			// Dis. Interrupt
	UART[Ux].TXwr++; UART[Ux].TXwr&=UartCntMask;
	UART[Ux].TXbuf[UART[Ux].TXwr]=value;

	if (UART[Ux].TXwr==UART[Ux].TXrd)  // Full?
	{
	  UART[Ux].TXrd++; UART[Ux].TXrd&=UartCntMask;
	  UART[Ux].TXerr|=1;  // Full, replace char
	}
	UART0_CTRLB|=(1<<UDRIE0);			// En. Interrupt
}


void UARTspeed(BYTE Ux, BYTE value)  // Set baudrate index
{
	switch(value)
	{
	  case 1: UBRR0=8;		// 8M/8/115200=9
	  	break;
	  case 2: UBRR0=103;		// 8M/8/9600=104
	  	break;
	  case 3: UBRR0=31;		// 8M/8/31250=32
	  	break;
	  default:
	  	break;
	}
	UART[Ux].Speed=value;
}



BYTE isUARTread(BYTE Ux)  // Data in RXbuf?
{
	BYTE value;

	value=(UART[Ux].RXrd!=UART[Ux].RXwr);
	return value;
}


BYTE isUARTreadT(BYTE Ux, BYTE Tm)  // Data in RXbuf? Timeout Tx100us
{
	while (UART[Ux].RXrd==UART[Ux].RXwr)
	{
	  Delay_us(100); Tm--;
	  if (!Tm) {return 0;}
	}
	return 1;
}


BYTE isUARTwrite(BYTE Ux)  // No full TXbuf?
{
	BYTE Delta;

	Delta=(UART[Ux].TXrd-UART[Ux].TXwr)&UartCntMask;
	return (BYTE)(Delta!=1); 
}


BYTE isUARTbusy(BYTE Ux)  // Data in TXbuf?
{
	BYTE value;

	value=(UART[Ux].TXrd!=UART[Ux].TXwr);
	return value;
}


void UARTinit(BYTE Ux)  // Clear buf
{
	UART[Ux].RXwr=0; UART[Ux].RXrd=0; UART[Ux].RXerr=0;
	UART[Ux].TXwr=0; UART[Ux].TXrd=0; UART[Ux].TXerr=0;
}


void UARTwriteS(BYTE Ux, const char *string)
{
  while (*string != '\0')
    {
	  if (!isUARTwrite(Ux)) {Delay_ms(1);}

 	  UARTwrite(Ux,*string);
      string++;
    }
}


void RXuart(BYTE Ux)  // wr++, buf[wr]=RX
{
  if (UART0_STATUS & (1<<RXC0))  // Read HW FIFO
  {
	UART[Ux].RXwr++; UART[Ux].RXwr&=UartCntMask;
	if (UART[Ux].RXwr==UART[Ux].RXrd)  // Full?
	{
	  UART[Ux].RXrd++; UART[Ux].RXrd&=UartCntMask;
	  UART[Ux].RXerr|=1;  // Full, replace char
	}
	UART[Ux].RXbuf[UART[Ux].RXwr]=UART0_DATA;
  }
}


void TXuart(BYTE Ux)  // rd++, TX=buf[rd], rd=wr? -> done, no interrupt
{
  if (UART0_STATUS & (1<<UDRE0))  // Write to HW FIFO
  {
	UART[Ux].TXrd++; UART[Ux].TXrd&=UartCntMask;
	UART0_DATA=UART[Ux].TXbuf[UART[Ux].TXrd];
	if (UART[Ux].TXrd==UART[Ux].TXwr)		// Done
	{
	  UART0_CTRLB&=~(1<<UDRIE0);			// Dis. Interrupt
	  //break;  // TX done
	}
  }
}


// UART0 RX interrupt handler
ISR(UART0_RECEIVE_INT)
{
	if ((UART0_STATUS & ((1<<FE0)|(1<<DOR0)))==0)  // Error?
	{
	  RXuart(UARTpss);
	}
}


// UART0 TX interrupt handler
ISR(UART0_TRANSMIT_INT)
{
 	 TXuart(UARTpss);
}
